#include "JLetterCounterObj.h"
#include "JInteger.h"
#include "JIntegerData.h"
#include "JStringData.h"
#include "JIntegerProperty.h"
#include "JIntegerListProperty.h"
#include "JStringProperty.h"
#include "JLinkObj.h"
#include "JBytesData.h"


char* theJLetterCounterObj = JLetterCounterObj().Register();


void JLetterCounterObj::writeContent(JOutputStream& os) {
  JAddObj::writeContent(os);
  putInteger(os, "numberColumns", numberColumns);
  putInteger(os, "valuesPerChunk", valuesPerChunk);
  putInteger(os, "onThreshhold", onThreshhold);
}

void JLetterCounterObj::readContent(JDictionary& dict) {
  JAddObj::readContent(dict);
  numberColumns = getInteger(dict, "numberColumns");
  valuesPerChunk = getInteger(dict, "valuesPerChunk");
  onThreshhold = getInteger(dict, "onThreshhold");
}

const char* JLetterCounterObj::className() const { return "JLetterCounterObj";}


JObject* JLetterCounterObj::clone() const { return new JLetterCounterObj(*this);}


// class constructor
JLetterCounterObj::JLetterCounterObj() {
  igm = 1;
  ogm = 1+1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1;
  numberColumns = 4;
  onThreshhold = 10;
  onColor = JColor::green;
  offColor = JColor::black;
  currentCount = 0;
  currentHistoryColumn = 0;
  totalCount = 0;
  initHistory();
}


void JLetterCounterObj::draw(JGraphics g, int x, int y, int w, int h) {
  JRect rect(x, y, w, h);
  drawText(g, JString("Letters"), rect);
}

void JLetterCounterObj::initCounts() {
	for ( int i = 0; i < 26; i++ ) {
		letterCounts[i] = 0;
	}
	for ( i = 0; i < 10; i++ ) {
		digitCounts[i] = 0;
	}
	currentCount = 0;
	//JComponent::debug( "initCounts" );
}

void JLetterCounterObj::initHistory() {
	for ( int i = 0; i < NUMBER_ROWS; i++ ) {
		for ( int j = 0; j < MAX_COLS; j++ ) {
			history[i][j] = 0;
		}
	}
}

#include <malloc.h>
#include <string.h>
#include <stdio.h>
char* newstrcpy( char* s ) {
	char* result = (char*)malloc( strlen( s ) + 1 );
	strcpy( result, s );
	return( result );
}
bool firstTime = true;
void JLetterCounterObj::paint(JGraphics g, double dx, double dy, JRegion& rgn, double scale) {
	//JComponent::debug( "Paint" );
    JRect rect = getIExtent(dx, dy, scale);
    int numberRows = 36;
    int yStart = rect.y;
	int yUnits = rect.height/numberRows;
	int xUnits = rect.width/numberColumns;
	int xStart = rect.x;
	int xErrorTotal = rect.width - xUnits * numberColumns;
	int yErrorTotal = rect.height - yUnits * numberRows;
	
	if ( firstTime ) {
		g.setJColor( JColor::black );
		g.fillRect(rect.x, rect.y, rect.width, rect.height);
	//char dbuf[200];
	//sprintf( dbuf, "x, y: %d, %d, w,h: %d, %d", rect.x, rect.y, rect.width, rect.height );
	//if ( firstTime ) JComponent::debug( newstrcpy( dbuf ));
 	}
   for ( int i = 0; i < numberRows; i++ ) {
		int height = yUnits;
		if ( yErrorTotal > 0 ) {
			yErrorTotal--;
			height++;
		}
		if ( i == ( numberRows - 1 )) {
			height += yErrorTotal;
		}
		int saveXerror = xErrorTotal;
		xStart = rect.x;
		for ( int j = 0; j < numberColumns; j++ ) {
			if ( history[i][j] ) {
				g.setJColor( onColor );
				//JComponent::debug( "it is On" );
			} else {
				g.setJColor( offColor );
				//JComponent::debug( "it is Off" );
			}
			int width = xUnits;
			if (xErrorTotal > 0) {
				xErrorTotal--;
				width++;
			} 
			if ( j == ( numberColumns - 1 )) {
				width += xErrorTotal;
			}
			g.fillRect( xStart, yStart, width, height );
			//sprintf( dbuf, "x, y: %d, %d, w,h: %d, %d", xStart + j*xUnits, yStart, xUnits, yUnits );
			//if ( firsttime ) JComponent::debug( newstrcpy( dbuf ));
			xStart += width;
		}
		yStart += height;
		xErrorTotal = saveXerror;
	}
	firstTime = false;
}


boolean JLetterCounterObj::inputAllowed(int n)
{ return !inputSet(n).last();}
JString JLetterCounterObj::inputTag(int n) {
  char* tag[] = {
    "text" };
  return( tag[n] );
}


int JLetterCounterObj::inputType( int n ) {
  if ( n == 0 ) return JStringData::id;
  return JIntegerData::id;
}
JString JLetterCounterObj::outputTag(int n) {
  char* tag[] = {
    "countA",
    "countB",
    "countC",
    "countD",
    "countE",
    "countF",
    "countG",
    "countH",
    "countI",
    "countJ",
    "countK",
    "countL",
    "countM",
    "countN",
    "countO",
    "countP",
    "countQ",
    "countR",
    "countS",
    "countT",
    "countU",
    "countV",
    "countW",
    "countX",
    "countY",
    "countZ",
    "count0",
    "count1",
    "count2",
    "count3",
    "count4",
    "count5",
    "count6",
    "count7",
    "count8",
    "count9",
  "total" };
  return( tag[n] );
}


JFRect JLetterCounterObj::inputArea(int n) {
  //if ( n < ( 1 )) {
    return leftArea( n, 0, 1, 0.1, 0.8 );
  //}
}


JFRect JLetterCounterObj::outputArea(int n) {
	if ( n < 36 )
  return rightArea(n, 0, 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1, 0.0, 1.0);
else
return bottomArea( n, 0, 1 );
}


int JLetterCounterObj::inputFace(int n) {
  if ( n < 1 ) {
    return LEFT;
  } else {
    return TOP;
  }
}


int JLetterCounterObj::outputFace(int n) {
	if ( n < 36 )
  return RIGHT;
	else
		return BOTTOM;
}


JFPoint JLetterCounterObj::inputPoint(int n, JLinkObj& link) {
  return leftPoint( n, link, 0, 1, 0.1, 0.8);
}


JFPoint JLetterCounterObj::outputPoint(int n, JLinkObj& link) {
	if ( n < 36 )
  return rightPoint(n, link, 0, 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1, 0.0, 1.0);
else
return bottomPoint( n, link, 0, 1 );
}


void JLetterCounterObj::access(int n, JLinkObj& link, const JDataType& data) {
  if (( n >= 0 ) && ( n < 26 )) {
    INT(data) = letterCounts[n];
  } else if ( n < 36 ) {
    INT(data) = digitCounts[n-26];
  } else {
	  INT(data) = totalCount;
  }
}

void JLetterCounterObj::broadcastCounts() {
	for ( int i = 0; i <= 36; i++ ) {
		broadcast( i );
	}
}

void JLetterCounterObj::shiftHistory() {
	for ( int i = 0; i < 26; i++ ) {
		history[i][currentHistoryColumn] = (letterCounts[i] > onThreshhold) ? 1 : 0;
	}
	for ( i = 0; i < 10; i++ ) {
		history[i+26][currentHistoryColumn] = (digitCounts[i] > onThreshhold) ? 1: 0;
	}
	currentHistoryColumn++;
	if ( currentHistoryColumn >= numberColumns ) {
		currentHistoryColumn = 0;
		firstTime = true;
	}
}

void JLetterCounterObj::process_AZ( int offset ) {
	letterCounts[offset] += 1;
}

void JLetterCounterObj::process_09( int offset ) {
	digitCounts[offset] += 1;
}

void JLetterCounterObj::processChar( char x ) {
	totalCount++;
	if (( x >= 'A' ) && ( x <= 'Z' )) {
		process_AZ( x - 'A' );
	} else if (( x >= 'a' ) && ( x <= 'z' )) {
		process_AZ( x - 'a' );
	} else if (( x >= '0' ) && ( x <= '9' )) {
		process_09( x - '0' );
	}
	currentCount++;
	if ( currentCount > valuesPerChunk ) {
		broadcastCounts();
		shiftHistory();
		repaint();
		initCounts();
	}
}

void JLetterCounterObj::processText() {
	char* scanner = text;
	while ( *scanner ) {
		processChar( *scanner++ );
	}
}

void JLetterCounterObj::engine(int n, JLinkObj& link) {
  switch (n) {
  case TEXT:
    link.access( JStringData( text ));
	processText();
    break;
  default:
    break;
  }
}


JArray JLetterCounterObj::getProperties() {
  JArray properties;
  properties.append(JIntegerProperty("numberColumns", numberColumns, 1, MAX_COLS));
  properties.append(JIntegerProperty("valuesPerChunk", valuesPerChunk, 1, 65535));
  properties.append(JIntegerProperty("onThreshhold", onThreshhold, 1, 65535));
  return properties;
}


boolean JLetterCounterObj::updateProperty(JProperty& prop) {
  if (prop.getName() == JString("numberColumns")) {
    numberColumns = ((JIntegerProperty*)&prop)->value;
    return true;
  } else if (prop.getName() == JString("valuesPerChunk")) {
    valuesPerChunk = ((JIntegerProperty*)&prop)->value;
    return true;
  } else if (prop.getName() == JString("onThreshhold")) {
    onThreshhold = ((JIntegerProperty*)&prop)->value;
    return true;
  }
  return false;
}


